package main

import (
	"bytes"
	"encoding/binary"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"os"
	"os/exec"
	"time"

	"golang.org/x/text/encoding/korean"
)

// Config 구조체 정의
type Config struct {
	ServerAddress string `json:"serverAddress"`
	RetryLimit    int    `json:"retryLimit"`
	RetryDelay    int    `json:"retryDelay"` // milliseconds
	LogFilePath   string `json:"logFilePath"`
	MaxLogSize    int64  `json:"maxLogSize"` // bytes
	ApiEndpoint   string `json:"apiEndpoint"`
}

// Packet 구조체 정의
type Packet struct {
	LogSize     byte
	LogType     byte
	LogDatetime [6]byte
	LogOn       byte
	LogAddress  [7]byte
	LogArea     [24]byte
	LogDevice   [24]byte
}

type ParsedData struct {
	LogSize     byte   `json:"logSize"`
	LogType     byte   `json:"logType"`
	LogDatetime string `json:"logDatetime"`
	LogOn       byte   `json:"logOn"`
	LogAddress  string `json:"logAddress"`
	LogArea     string `json:"logArea"`
	LogDevice   string `json:"logDevice"`
}

const readTimeout = 1 * time.Minute // 응답 타임아웃 설정 (1분)

func main() {
	// 설정 파일 읽기
	config, err := loadConfig("config.json")
	if err != nil {
		log.Fatalf("Failed to load config: %v", err)
	}

	// 로그 파일 설정
	logFile, err := os.OpenFile(config.LogFilePath, os.O_APPEND|os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		log.Fatalf("Failed to open log file: %v", err)
	}
	defer logFile.Close()
	log.SetOutput(io.MultiWriter(os.Stdout, logFile))

	var conn net.Conn

	for {
		// 재시도 로직
		for attempts := 0; attempts < config.RetryLimit; attempts++ {
			conn, err = net.Dial("tcp", config.ServerAddress)
			if err == nil {
				log.Println("Connected to server")
				break
			}
			log.Printf("Failed to connect to server (attempt %d/%d): %v", attempts+1, config.RetryLimit, err)
			time.Sleep(time.Duration(config.RetryDelay) * time.Millisecond)
		}

		if err != nil {
			log.Printf("Could not connect to server after %d attempts: %v", config.RetryLimit, err)
			rebootSystem()
			return
		}

		// 서버와 지속적인 연결 유지
		for {
			conn.SetReadDeadline(time.Now().Add(readTimeout))

			packet := &Packet{}
			err = binary.Read(conn, binary.BigEndian, packet)
			if err != nil {
				log.Printf("Read error: %v", err)
				conn.Close()
				break
			}

			// 각 필드 파싱 및 출력
			parsedData := ParsedData{
				LogSize:     packet.LogSize,
				LogType:     packet.LogType,
				LogDatetime: parseDatetime(packet.LogDatetime[:]),
				LogOn:       packet.LogOn,
				LogAddress:  string(bytes.Trim(packet.LogAddress[:], "\x00")),
				LogArea:     decodeEUC_KR(packet.LogArea[:]),
				LogDevice:   decodeEUC_KR(packet.LogDevice[:]),
			}

			log.Printf("Parsed Data: %+v\n", parsedData)

			// 데이터 전송
			err = sendToApi(config.ApiEndpoint, parsedData)
			if err != nil {
				log.Printf("Failed to send data to API: %v", err)
			}

			// 로그 파일 크기 체크 및 리셋
			if err := checkLogSizeAndReset(config.LogFilePath, config.MaxLogSize); err != nil {
				log.Fatalf("Failed to check log size: %v", err)
			}
		}
	}
}

// 설정 파일을 읽어오는 함수
func loadConfig(filename string) (*Config, error) {
	data, err := ioutil.ReadFile(filename)
	if err != nil {
		return nil, err
	}

	var config Config
	err = json.Unmarshal(data, &config)
	if err != nil {
		return nil, err
	}

	return &config, nil
}

// EUC-KR 인코딩을 UTF-8로 디코딩하는 함수
func decodeEUC_KR(data []byte) string {
	dec := korean.EUCKR.NewDecoder()
	decoded, err := dec.Bytes(data)
	if err != nil {
		log.Fatalf("Failed to decode EUC-KR: %v", err)
	}
	return string(decoded)
}

// 시스템을 재부팅하는 함수
func rebootSystem() {
	log.Println("Rebooting the system...")
	cmd := exec.Command("reboot")
	err := cmd.Run()
	if err != nil {
		log.Fatalf("Failed to reboot the system: %v", err)
	}
}

// 로그 파일 크기 체크 및 리셋하는 함수
func checkLogSizeAndReset(logFilePath string, maxSize int64) error {
	fileInfo, err := os.Stat(logFilePath)
	if err != nil {
		return err
	}

	if fileInfo.Size() >= maxSize {
		log.Println("Log file size exceeded, resetting log file")
		return resetLogFile(logFilePath)
	}

	return nil
}

// 로그 파일을 리셋하는 함수
func resetLogFile(logFilePath string) error {
	// 기존 파일 닫기
	logFile, err := os.OpenFile(logFilePath, os.O_TRUNC|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}
	logFile.Close()

	// 새로운 로그 파일 열기
	logFile, err = os.OpenFile(logFilePath, os.O_APPEND|os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}

	log.SetOutput(io.MultiWriter(os.Stdout, logFile))
	return nil
}

// 데이터를 REST API로 전송하는 함수
func sendToApi(apiEndpoint string, data ParsedData) error {
	jsonData, err := json.Marshal(data)
	if err != nil {
		return err
	}

	req, err := http.NewRequest("POST", apiEndpoint, bytes.NewBuffer(jsonData))
	if err != nil {
		return err
	}

	req.Header.Set("Content-Type", "application/json")

	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return err
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("API request failed with status: %s", resp.Status)
	}

	return nil
}

// 로그 발생 시각을 문자열로 변환하는 함수
func parseDatetime(data []byte) string {
	return fmt.Sprintf("%02x-%02x-%02x %02x:%02x:%02x", data[0], data[1], data[2], data[3], data[4], data[5])
}

